from pylons import config
from sqlalchemy import Column, MetaData, Table, types, ForeignKey
from sqlalchemy.orm import mapper, relation
from sqlalchemy.orm import scoped_session, sessionmaker

class BaseBlock(object):
#     __slots__ = ('x', 'y', 'top', 'left', 'right', 'bottom')

    def __init__(self, sides, position=None):
        self.sides = sides

        if position is not None:
            self.x, self.y = position

    def _get_sides(self):
        return (self.top, self.right, self.bottom, self.left)

    def _set_sides(self, sides):
        self.top, self.right, self.bottom, self.left = sides

    sides = property(_get_sides, _set_sides)

    def __str__(self):
        args = (self.x, self.y) + self.sides
        return 'Block at (%s, %s), top: %s, right: %s, bottom: %s, left: %s' % args

    def rotate(self, degrees):
        offset = degrees / 90
        self.sides = self.rotated_sides(offset)

    def rotated_sides(self, offset):
        return self.sides[-offset:] + self.sides[:-offset]


class Block(BaseBlock):
    pass

class BaseGameField(object):
#     __slots__ = ('block_factory', 'blocks', 'size_x', 'size_y')

    block_factory = BaseBlock
    
    def __init__(self, size):
        self.blocks = []
        self.initialize_field(size)

    def initialize_field(self, size):
        self.size_x, self.size_y = size

        for y in xrange(self.size_y):
            for x in xrange(self.size_x):
                self.blocks.append(self.block_factory((None, ) * 4, (x, y)))

    @property
    def _field(self):
        for y in xrange(self.size_y):
            row = []
            for x in xrange(self.size_x):
                row.append(self.getblock(x, y))
            yield row

    def getblock(self, x, y):
        for b in self.blocks:
            if b.x == x and b.y == y:
                return b
        raise KeyError('No block found at x: %s, y: %s' % (x, y))

    def setblock(self, block, x, y):
        for b in self.blocks:
            if b.x == x and b.y == y:
                b.sides = block.sides

    def is_solved(self, last_block=None):
        def sides_match(a, b):
            # first test for jokers
            if a == 'j' or b == 'j':
                return True
            return a == b
        
        if last_block is not None:
            block_x = last_block.x
            block_y = last_block.y

            for block_side, other_side, add_x, add_y in ((1, 3, 1, 0), 
                                                         (2, 0, 0, 1), 
                                                         (3, 1, -1, 0), 
                                                         (0, 2, 0, -1)):
                try:
                    if not sides_match(last_block.sides[block_side], 
                                       self.getblock(block_x + add_x, 
                                                     block_y + add_y).sides[other_side]):
                        return False
                except KeyError:
                    pass

        max_x, max_y = self.size_x, self.size_y

        for y, row in enumerate(self._field):
            for x, block in enumerate(row):
                # Check if we are not at the last block
                if x < (max_x - 1):
                    if not sides_match(block.right, row[x + 1].left):
                        return False
                if y < (max_y - 1):
                    if not sides_match(block.bottom, self.getblock(x, y + 1).top):
                        return False
        return True
                    
    def __str__(self):
        text = []
        format = lambda s: str(s).rjust(2)

        for i, row in enumerate(self._field):
            text.append(
                '  '.join(['  %s  ' % format(block.top) for block in row]))
            text.append(
                '  '.join(['%s  %s' % (format(block.left), format(block.right)) for block in row]))
            text.append(
                '  '.join(['  %s  ' % format(block.bottom) for block in row]))
            if i < self.size_y - 1:
                text.append('')
        return '\n'.join(text)

class GameField(BaseGameField):
    block_factory = Block

    def __init__(self, size):
        self.initialize_field(size)

    def light_weight_copy(self):
        field = BaseGameField((self.size_x, self.size_y))
        for block in self.blocks:
            field.setblock(block, block.x, block.y)
        return field
    

class Game(object):
    def __init__(self, field):
        self.field = field

Session = scoped_session(sessionmaker(autoflush=True,
                         transactional=True, bind=config['pylons.g'].sa_engine))

metadata = MetaData()

blocks_table = Table(
    'blocks', metadata,
    Column('id', types.Integer(), primary_key=True),
    Column('x', types.Integer()),
    Column('y', types.Integer()),
    Column('top', types.Integer()),
    Column('left', types.Integer()),
    Column('bottom', types.Integer()),
    Column('right', types.Integer()),
    Column('field_id', types.Integer(), ForeignKey('gamefields.id')),
)
mapper(Block, blocks_table)

gamefields_table = Table(
    'gamefields', metadata,
    Column('id', types.Integer(), primary_key=True),
    Column('size_x', types.Integer()),
    Column('size_y', types.Integer()),
)
mapper(GameField, gamefields_table, properties={    
        'blocks': relation(Block, backref='field'),
        })

games_table = Table(
    'games', metadata,
    Column('id', types.Integer(), primary_key=True),
    Column('steps_made', types.Integer(), default=0),
    Column('maximum_steps', types.Integer(), default=0),
    Column('field_id', types.Integer(), ForeignKey('gamefields.id')),
)

mapper(Game, games_table, properties={    
        'field':relation(GameField, backref='game')})
